/*
btopen.c Version 0.1.0  Btree open database
Copyright (C) 2003  dondalah@ripco.com (Dondalah)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to:

	Free Software Foundation, Inc.
	59 Temple Place - Suite 330
	Boston, MA  02111-1307, USA.
*/

#include <bthash.h>

#define BTOPEN 200

btreefmt *btopen(dbname,rwflag,lock,reason)
char *dbname;
int rwflag;
int lock;
int *reason;
   {
   int rslt;
   int rdlen;
   int hndl;
   unsigned char *p,*q;
   btreefmt *bthdr;
   if (rwflag == UPDT)
      hndl = open(dbname,O_RDWR);
   else if (rwflag == READ)
      hndl = open(dbname,O_RDONLY);
   else
      {
      fprintf(stderr,"btopen: invalid read/write flag\n");
      *reason = BTREE_ERR + BTOPEN + 1;
      return(NULL);
      } /* rwflag err */
   if (hndl < 0)
      {
      fprintf(stderr,"btopen: %s: open ", dbname);
      if (rwflag == READ)
         fprintf(stderr,"read-only\n");
      else if (rwflag == UPDT)
         fprintf(stderr,"read/write\n");
      perror("Open error");
      *reason = SYS_ERR + errno;
      return(NULL);
      } /* if open error */
   if (lock == LOCK)
      {
      rslt = lockf(hndl,F_TLOCK,0);
      if (rslt < 0)
         {
         fprintf(stderr,"btopen: lock "
	    "error on %s\n", dbname);
         perror("Lock error");
         rslt = close(hndl);
         *reason = SYS_ERR + errno;
         return(NULL);
         } /* if lock failed */
      } /* if lock */
   else if (lock == TSTLOCK)
      {
      rslt = lockf(hndl,F_TEST,0);
      if (rslt < 0)
         {
         fprintf(stderr,"btopen: lock test "
	    "error on %s\n", dbname);
         perror("Lock test error");
         rslt = close(hndl);
         *reason = SYS_ERR + errno;
         return(NULL);
         } /* if lock test failed */
      } /* if lock test */
   else if (lock == NOLOCK);
   else
      {
      fprintf(stderr,"btopen: invalid lock command\n");
      rslt = close(hndl);
      *reason = BTREE_ERR + BTOPEN + 2;
      return(NULL);
      } /* lock cmd error */
   bthdr = (btreefmt *) malloc(ROOT+ROOT);
   if (bthdr == NULL)
      {
      fprintf(stderr,"btopen: out of memory "
	 "allocating bthdr\n");
      rslt = lockf(hndl,F_ULOCK,0);
      rslt = close(hndl);
      *reason = BTREE_ERR + BTOPEN + 3;
      return(NULL);
      } /* out of mem */
   rdlen = read(hndl,bthdr,ROOT);
   if (rdlen != ROOT)
      {
      fprintf(stderr,"btopen: dbname %s\n", dbname);
      perror("btopen: error reading header");
      rslt = lockf(hndl,F_ULOCK,0);
      rslt = close(hndl);
      free(bthdr);
      *reason = BTREE_ERR + BTOPEN + 4;
      return(NULL);
      } /* read err */
   if (strncmp(bthdr->signature,
      BT_SIGNATURE,sizeof(BT_SIGNATURE)))
      {
      fprintf(stderr,"btopen: dbname %s\n", dbname);
      fprintf(stderr,"btopen: error "
	 "opening non-btree file\n");
      rslt = lockf(hndl,F_ULOCK,0);
      rslt = close(hndl);
      free(bthdr);
      *reason = BTREE_ERR + BTOPEN + 5;
      return(NULL);
      } /* invalid signature */
   p = (unsigned char *) bthdr + ROOT;
   q = (unsigned char *) p + ROOT;
   while (p < q) *p++ = '\0';
   bthdr->hndl = hndl;
   bthdr->lock = lock;

   /* nodesz = records per node */
   bthdr->rcdnode = bthdr->rcdhdr + bthdr->rcdlen;
   bthdr->blknode = bthdr->blkhdr + bthdr->blksz;
   bthdr->midofst = bthdr->blkhdr
      + (bthdr->rcdnode * (bthdr->nodesz >> 1));

   bthdr->head = (stkfmt *) btstkini(bthdr);
   bthdr->tail = (stkfmt *) btstkini(bthdr);
   bthdr->head->prev = bthdr->head;
   bthdr->head->next = bthdr->tail;
   bthdr->tail->prev = bthdr->head;
   bthdr->tail->next = bthdr->tail;
   bthdr->blk = (blkfmt *) malloc(bthdr->blksz+16);
   if (bthdr->blk == NULL)
      {
      fprintf(stderr,"btopen: out of memory "
	 "allocating bthdr->buf\n");
      rslt = lockf(hndl,F_ULOCK,0);
      btclose(bthdr);
      free(bthdr);
      *reason = BTREE_ERR + BTOPEN + 6;
      return(NULL);
      } /* out of mem */
   bthdr->blkleft = (blkfmt *) malloc(bthdr->blksz+16);
   if (bthdr->blkleft == NULL)
      {
      fprintf(stderr,"btopen: out of memory "
	 "allocating bthdr->buf2\n");
      rslt = lockf(hndl,F_ULOCK,0);
      btclose(bthdr);
      free(bthdr->blk);
      free(bthdr);
      *reason = BTREE_ERR + BTOPEN + 7;
      return(NULL);
      } /* out of mem */
   p = (unsigned char *) bthdr->blkleft;
   q = (unsigned char *) p + bthdr->blksz;
   while (p < q) *p++ = '\0';
   bthdr->blkrght = (blkfmt *) malloc(bthdr->blksz+16);
   if (bthdr->blkrght == NULL)
      {
      fprintf(stderr,"btopen: out of memory "
	 "allocating bthdr->buf3\n");
      rslt = lockf(hndl,F_ULOCK,0);
      btclose(bthdr);
      free(bthdr->blk);
      free(bthdr->blkleft);
      free(bthdr);
      *reason = BTREE_ERR + BTOPEN + 8;
      return(NULL);
      } /* out of mem */
   p = (unsigned char *) bthdr->blkrght;
   q = (unsigned char *) p + bthdr->blksz;
   while (p < q) *p++ = '\0';
   bthdr->node = (bt_nodefmt *) malloc(bthdr->rcdnode+16);
   if (bthdr->node == NULL)
      {
      fprintf(stderr,"btopen: out of memory "
	 "allocating bthdr->node\n");
      rslt = lockf(hndl,F_ULOCK,0);
      btclose(bthdr);
      free(bthdr->blk);
      free(bthdr->blkleft);
      free(bthdr->blkrght);
      free(bthdr);
      *reason = BTREE_ERR + BTOPEN + 9;
      return(NULL);
      } /* out of mem */
   p = (unsigned char *) bthdr->node;
   q = (unsigned char *) p + bthdr->rcdnode;
   while (p < q) *p++ = '\0';
   strcpy(bthdr->dbname,dbname);
   return(bthdr);
   } /* btopen */
